/*
  Copyright (c) 2009 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <stdio.h>
#include <spu_mfcio.h>
#include <spu_intrinsics.h>

#include "mt_mine.h"

/* Period parameters */
#define M 397
#define N 624
#define MATRIX_A 0x9908b0dfUL   /* constant vector a */
#define UPPER_MASK 0x80000000UL /* most significant w-r bits */
#define LOWER_MASK 0x7fffffffUL /* least significant r bits */

static unsigned long mt[N] __attribute__((aligned(16))); 
static int mti = N + 1; /* mti==N+1 means mt[N] is not initialized */

static vec_uint4 xmt[32][5] __attribute__((aligned(16)));


/* initializes mt[N] with a seed */
void init_genrand_mine(unsigned long s)
{
    mt[0]= s & 0xffffffffUL;
    for (mti=1; mti<N; mti++) {
        mt[mti] =
            (1812433253UL * (mt[mti-1] ^ (mt[mti-1] >> 30)) + mti);
        /* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
        /* In the previous versions, MSBs of the seed affect   */
        /* only MSBs of the array mt[].                        */
        /* 2002/01/09 modified by Makoto Matsumoto             */
        mt[mti] &= 0xffffffffUL;
        /* for >32 bit machines */
    }
}


#define XOR3(A, B, C) spu_xor(B, spu_xor(C, A))

#define UPDATE_TABLE(UNIT, BIT)                                         \
    if (UNIT == 0) {                                                    \
        vec_uint4 a = xmt[BIT][3];                                      \
        vec_uint4 b = xmt[BIT][4];                                      \
        far1 = spu_rlqwbytebc(spu_rlqw(spu_sel(a, b, vsel13), 13), 13); \
        if (BIT == 0)                                                   \
            x1[BIT] = spu_xor(far1, y31);                               \
        else {                                                          \
            if ((MATRIX_A >> (31 - BIT)) & 1)                           \
                x1[BIT] = XOR3(far1, y_cur, y31);                       \
            else                                                        \
                x1[BIT] = spu_xor(far1, y_cur);                         \
            y_cur = y_next;                                             \
        }                                                               \
    } else if (UNIT == 1) {                                             \
        vec_uint4 a = xmt[BIT][4];                                      \
        vec_uint4 b = xmt[BIT][0];                                      \
        far1 = spu_rlqwbytebc(spu_rlqw(spu_sel(spu_rlmaskqwbyte(a, -2), \
                                               b, vsel29),              \
                                       29), 29);                        \
        if (BIT == 0)                                                   \
            x1[BIT] = spu_xor(far1, y31);                               \
        else {                                                          \
            if ((MATRIX_A >> (31 - BIT)) & 1)                           \
                x1[BIT] = XOR3(far1, y_cur, y31);                       \
            else                                                        \
                x1[BIT] = spu_xor(far1, y_cur);                         \
            y_cur = y_next;                                             \
        }                                                               \
    } else {                                                            \
        vec_uint4 a = xmt[BIT][UNIT - 2];                               \
        vec_uint4 b = xmt[BIT][UNIT - 1];                               \
        far1 = spu_rlqwbytebc(spu_rlqw(spu_sel(a, b, vsel29), 29), 29); \
        if (BIT == 0)                                                   \
            x1[BIT] = spu_xor(far1, y31);                               \
        else {                                                          \
            if ((MATRIX_A >> (31 - BIT)) & 1)                           \
                x1[BIT] = XOR3(far1, y_cur, y31);                       \
            else                                                        \
                x1[BIT] = spu_xor(far1, y_cur);                         \
            y_cur = y_next;                                             \
        }                                                               \
    } 



#define MAKE_Y(UNIT, BIT)                                               \
    if (BIT == 0)                                                       \
        y_cur = xmt[0][UNIT];                                           \
    else if (UNIT == 4)                                                 \
        if (BIT == 31)                                                  \
            y31 = spu_slqw(spu_sel(xmt[31][4],                          \
                                   spu_rlqwbyte(xmt[31][0], 2),         \
                                   vsel112), 1);                        \
        else                                                            \
            y_next = spu_slqw(spu_sel(xmt[BIT][4],                      \
                                      spu_rlqwbyte(xmt[BIT][0], 2),     \
                                      vsel112), 1);                     \
    else                                                                \
        if (BIT == 31)                                                  \
            y31 = spu_rlqw(spu_sel(xmt[31][UNIT], xmt[31][UNIT + 1],    \
                                   vsel1), 1);                          \
        else                                                            \
            y_next = spu_rlqw(spu_sel(xmt[BIT][UNIT],                   \
                                      xmt[BIT][UNIT + 1], vsel1), 1);



# define PRE_ACCUMULATE(UNIT, WAY) {                                    \
        vec_uchar16 a3, a2, a1, a0;                                     \
        if (UNIT == 4) {                                                \
            a3 = spu_slqw(spu_cntb((vec_uchar16)spu_rlmaskqwbyte(x40, -2)), \
                          3);                                           \
            a1 = spu_slqw(spu_cntb((vec_uchar16)spu_rlmaskqwbyte(x42, -2)), \
                          1);                                           \
            a3 = (vec_uchar16)spu_add((vec_ushort8)a1, (vec_ushort8)a3); \
                                                                        \
            a2 = spu_slqw(spu_cntb((vec_uchar16)spu_rlmaskqwbyte(x41, -2)), \
                          2);                                           \
            a0 = spu_cntb((vec_uchar16)spu_rlmaskqwbyte(x43, -2));      \
                                                                        \
            a2 = (vec_uchar16)spu_add((vec_ushort8)a0, (vec_ushort8)a2); \
            vcsum[WAY]                                           \
                = (vec_uchar16)spu_add((vec_ushort8)a3,                \
                                       (vec_ushort8)a2);              \
        } else {                                                        \
            a3 = spu_slqw(spu_cntb((vec_uchar16)x40), 3);               \
            a1 = spu_slqw(spu_cntb((vec_uchar16)x42), 1);               \
            a3 = (vec_uchar16)spu_add((vec_ushort8)a1, (vec_ushort8)a3); \
            a2 = spu_slqw(spu_cntb((vec_uchar16)x41), 2);               \
            a0 = spu_cntb((vec_uchar16)x43);                            \
            a2 = (vec_uchar16)spu_add((vec_ushort8)a0, (vec_ushort8)a2); \
            vcsum[WAY] = (vec_uchar16)spu_add((vec_ushort8)a3,          \
                                                   (vec_ushort8)a2);    \
        }                                                               \
    }
    


# define PRE_ACCUMULATE_LAST(UNIT, WAY, MASK) {                         \
        vec_uchar16 a3, a2, a1, a0;                                     \
        if (UNIT == 4) {                                                \
            a3 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x40, MASK)),    \
                          3);                                           \
            a1 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x42, MASK)),    \
                          1);                                           \
            a3 = (vec_uchar16)spu_add((vec_ushort8)a1, (vec_ushort8)a3); \
            a2 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x41, MASK)),    \
                          2);                                           \
            a0 = spu_cntb((vec_uchar16)spu_and(x43, MASK));             \
                                                                        \
            a2 = (vec_uchar16)spu_add((vec_ushort8)a0, (vec_ushort8)a2); \
            vcsum[WAY]                                           \
                = (vec_uchar16)spu_add((vec_ushort8)a3, (vec_ushort8)a2); \
        } else {                                                        \
            a3 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x40, MASK)),    \
                          3);                                           \
            a1 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x42, MASK)),    \
                          1);                                           \
            a3 = (vec_uchar16)spu_add((vec_ushort8)a1, (vec_ushort8)a3); \
            a2 = spu_slqw(spu_cntb((vec_uchar16)spu_and(x41, MASK)),    \
                          2);                                           \
            a0 = spu_cntb((vec_uchar16)spu_and(x43, MASK));             \
                                                                        \
            a2 = (vec_uchar16)spu_add((vec_ushort8)a0, (vec_ushort8)a2); \
            vcsum[WAY]                                           \
                = (vec_uchar16)spu_add((vec_ushort8)a3, (vec_ushort8)a2); \
        }                                                               \
    }



# define GEN_RAND_CALC(UNIT) {                                          \
        x0_7 = spu_xor(x1[0], x1[7]);                                   \
        x1_5 = spu_xor(x1[1], x1[5]);                                   \
        x3_10 = spu_xor(x1[3], x1[10]);                                 \
        x6_10 = spu_xor(x1[6], x1[10]);                                 \
        x0_11 = spu_xor(x1[0], x1[11]);                                 \
        x7_11 = spu_xor(x1[7], x1[11]);                                 \
        x1_12 = spu_xor(x1[1], x1[12]);                                 \
        x8_12 = spu_xor(x1[8], x1[12]);  /* */                          \
        x3_14 = spu_xor(x1[3], x1[14]);                                 \
        x4_15 = spu_xor(x1[4], x1[15]);                                 \
        x5_16 = spu_xor(x1[5], x1[16]);                                 \
        x1_5_16 = spu_xor(x1_5, x1[16]);                                \
        x6_17 = spu_xor(x1[6], x1[17]);                                 \
        x0_18 = spu_xor(x1[0], x1[18]);                                 \
        x8_19 = spu_xor(x1[8], x1[19]);                                 \
        x9_20 = spu_xor(x1[9], x1[20]);                                 \
        x17_21 = spu_xor(x1[17], x1[21]);                               \
        x7_11_22 = spu_xor(x7_11, x1[22]);                              \
        x8_12_23 = spu_xor(x8_12, x1[23]);                              \
        x13_24 = spu_xor(x1[13], x1[24]);                               \
        x3_25 = spu_xor(x1[3], x1[25]);                                 \
        x15_26 = spu_xor(x1[15], x1[26]);                               \
        x3_28 = spu_xor(x1[3], x1[28]);                                 \
        x20_28 = spu_xor(x1[20], x1[28]);                               \
        x6_10_28 = spu_xor(x6_10, x1[28]);                              \
        x2_9_13_17_28 = XOR3(x1[28], spu_xor(x1[9], x1[13]),            \
                             spu_xor(x1[2], x1[17]));                   \
                                                                        \
        x22_29 = spu_xor(x1[22], x1[29]);                               \
        x18_29 = spu_xor(x1[18], x1[29]);                               \
        x8_30 = spu_xor(x1[8], x1[30]);                                 \
        x20_31 = spu_xor(x1[20], x1[31]);                               \
        x28_31 = spu_xor(x1[28], x1[31]);                               \
        x9_13_20_24_31 = XOR3(x1[31], x9_20, x13_24);                   \
        x1_5_9_20 = spu_xor(x1_5, x9_20);                               \
        x6_13_17_24 = spu_xor(x6_17, x13_24);                           \
        x8_15_19_26 = spu_xor(x15_26, x8_19);                           \
        x3_14_18_29 = spu_xor(x3_14, x18_29);                           \
        x0_4_15_18 = spu_xor(x0_18, x4_15);                             \
        x6_10_17_21_28 = spu_xor(x17_21, x6_10_28);                     \
                                                                        \
                                                                        \
                                                                        \
        x40 = spu_xor(x1_12, x8_19);                                    \
        x41 = spu_xor(x2_9_13_17_28, x1[20]);                           \
        x42 = x3_14_18_29;                                              \
        x43 = x4_15;                                                    \
        xmt[12][UNIT] = x1[12];                                         \
        xmt[13][UNIT] = x1[13];                                         \
        xmt[14][UNIT] = x1[14];                                         \
        xmt[15][UNIT] = x1[15];                                         \
        PRE_ACCUMULATE(UNIT, 0);                                        \
                                                                        \
        x40 = x6_10_28;                                                 \
        x41 = spu_xor(x0_11, x18_29);                                   \
        x42 = spu_xor(x8_30, x1_12);                                    \
        x43 = spu_xor(x2_9_13_17_28, x1[31]);                           \
        xmt[28][UNIT] = x1[28];                                         \
        xmt[29][UNIT] = x1[29];                                         \
        xmt[30][UNIT] = x1[30];                                         \
        xmt[31][UNIT] = x1[31];                                         \
        PRE_ACCUMULATE(UNIT, 1);                                       \
                                                                        \
        vdsum[3] = spu_add(vdsum[3], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
                                                                        \
        x40 = x8_12_23;                                                 \
        x41 = x9_13_20_24_31;                                           \
        x42 = spu_xor(x6_10, x1[17]);                                   \
        x43 = x0_11;                                                    \
        xmt[8][UNIT] = x1[8];                                           \
        xmt[9][UNIT] = x1[9];                                           \
        xmt[10][UNIT] = x1[10];                                         \
        xmt[11][UNIT] = x1[11];                                         \
        PRE_ACCUMULATE(UNIT, 0);                                       \
                                                                        \
        x40 = XOR3(x13_24, x20_31, x6_10_17_21_28);                     \
        x41 = XOR3(x7_11_22, x18_29, x3_25);                            \
        x42 = spu_xor(x15_26, x8_12_23);                                \
        x43 = XOR3(x1[16], x9_13_20_24_31, x1[27]);                     \
        xmt[24][UNIT] = x1[24];                                         \
        xmt[25][UNIT] = x1[25];                                         \
        xmt[26][UNIT] = x1[26];                                         \
        xmt[27][UNIT] = x1[27];                                         \
        PRE_ACCUMULATE(UNIT, 1);                                       \
                                                                        \
        vdsum[2] = spu_add(vdsum[2], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
        x40 = spu_xor(x0_7, x4_15);                                     \
        x41 = x1_5_16;                                                  \
        x42 = spu_xor(x6_13_17_24, x1[2]);                              \
        x43 = x3_10;                                                    \
        xmt[0][UNIT] = x1[0];                                           \
        xmt[1][UNIT] = x1[1];                                           \
        xmt[2][UNIT] = x1[2];                                           \
        xmt[3][UNIT] = x1[3];                                           \
        PRE_ACCUMULATE(UNIT, 0);                                        \
                                                                        \
        x40 = x5_16;                                                    \
        x41 = x6_13_17_24;                                              \
        x42 = x0_4_15_18;                                               \
        x43 = spu_xor(x8_15_19_26, x1_5_16);                            \
        xmt[16][UNIT] = x1[16];                                         \
        xmt[17][UNIT] = x1[17];                                         \
        xmt[18][UNIT] = x1[18];                                         \
        xmt[19][UNIT] = x1[19];                                         \
        PRE_ACCUMULATE(UNIT, 1);                                       \
                                                                        \
        vdsum[0] = spu_add(vdsum[0], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
        x40 = spu_xor(x0_11, XOR3(x1[26], x8_19, x4_15));               \
        x41 = spu_xor(x1_5_9_20, x1[12]);                               \
        x42 = x6_10_17_21_28;                                           \
        x43 = spu_xor(x7_11_22, x3_14_18_29);                           \
        xmt[4][UNIT] = x1[4];                                           \
        xmt[5][UNIT] = x1[5];                                           \
        xmt[6][UNIT] = x1[6];                                           \
        xmt[7][UNIT] = x1[7];                                           \
        PRE_ACCUMULATE(UNIT, 0);                                        \
                                                                        \
        x40 = XOR3(x1[2], x9_20, x6_13_17_24);                          \
        x41 = spu_xor(x3_28, x17_21);                                   \
        x42 = XOR3(XOR3(x1[26], x8_19, x4_15), x0_18, x22_29);          \
        x43 = spu_xor(x1_5_9_20, x1[23]);                               \
        xmt[20][UNIT] = x1[20];                                         \
        xmt[21][UNIT] = x1[21];                                         \
        xmt[22][UNIT] = x1[22];                                         \
        xmt[23][UNIT] = x1[23];                                         \
        PRE_ACCUMULATE(UNIT, 1);                                       \
                                                                        \
        vdsum[1] = spu_add(vdsum[1], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
                                                                        \
    }


# define GEN_RAND_CALC_LAST(UNIT, MASK) {                               \
        x0_7 = spu_xor(x1[0], x1[7]);                                   \
        x1_5 = spu_xor(x1[1], x1[5]);                                   \
        x3_10 = spu_xor(x1[3], x1[10]);                                 \
        x6_10 = spu_xor(x1[6], x1[10]);                                 \
        x0_11 = spu_xor(x1[0], x1[11]);                                 \
        x7_11 = spu_xor(x1[7], x1[11]);                                 \
        x1_12 = spu_xor(x1[1], x1[12]);                                 \
        x8_12 = spu_xor(x1[8], x1[12]);  /* */                          \
        x3_14 = spu_xor(x1[3], x1[14]);                                 \
        x4_15 = spu_xor(x1[4], x1[15]);                                 \
        x5_16 = spu_xor(x1[5], x1[16]);                                 \
        x1_5_16 = spu_xor(x1_5, x1[16]);                                \
        x6_17 = spu_xor(x1[6], x1[17]);                                 \
        x0_18 = spu_xor(x1[0], x1[18]);                                 \
        x8_19 = spu_xor(x1[8], x1[19]);                                 \
        x9_20 = spu_xor(x1[9], x1[20]);                                 \
        x17_21 = spu_xor(x1[17], x1[21]);                               \
        x7_11_22 = spu_xor(x7_11, x1[22]);                              \
        x8_12_23 = spu_xor(x8_12, x1[23]);                              \
        x13_24 = spu_xor(x1[13], x1[24]);                               \
        x3_25 = spu_xor(x1[3], x1[25]);                                 \
        x15_26 = spu_xor(x1[15], x1[26]);                               \
        x3_28 = spu_xor(x1[3], x1[28]);                                 \
        x20_28 = spu_xor(x1[20], x1[28]);                               \
        x6_10_28 = spu_xor(x6_10, x1[28]);                              \
        x2_9_13_17_28 = XOR3(x1[28], spu_xor(x1[9], x1[13]),            \
                             spu_xor(x1[2], x1[17]));                   \
                                                                        \
        x22_29 = spu_xor(x1[22], x1[29]);                               \
        x18_29 = spu_xor(x1[18], x1[29]);                               \
        x8_30 = spu_xor(x1[8], x1[30]);                                 \
        x20_31 = spu_xor(x1[20], x1[31]);                               \
        x28_31 = spu_xor(x1[28], x1[31]);                               \
        x9_13_20_24_31 = XOR3(x1[31], x9_20, x13_24);                   \
        x1_5_9_20 = spu_xor(x1_5, x9_20);                               \
        x6_13_17_24 = spu_xor(x6_17, x13_24);                           \
        x8_15_19_26 = spu_xor(x15_26, x8_19);                           \
        x3_14_18_29 = spu_xor(x3_14, x18_29);                           \
        x0_4_15_18 = spu_xor(x0_18, x4_15);                             \
        x6_10_17_21_28 = spu_xor(x17_21, x6_10_28);                     \
                                                                        \
                                                                        \
        x40 = spu_xor(x0_7, x4_15);                                     \
        x41 = x1_5_16;                                                  \
        x42 = spu_xor(x6_13_17_24, x1[2]);                              \
        x43 = x3_10;                                                    \
        xmt[0][UNIT] = x1[0];                                           \
        xmt[1][UNIT] = x1[1];                                           \
        xmt[2][UNIT] = x1[2];                                           \
        xmt[3][UNIT] = x1[3];                                           \
        PRE_ACCUMULATE_LAST(UNIT, 0, MASK);                             \
                                                                        \
        x40 = x5_16;                                                    \
        x41 = x6_13_17_24;                                              \
        x42 = x0_4_15_18;                                               \
        x43 = spu_xor(x8_15_19_26, x1_5_16);                            \
        if (UNIT != 4) {                                                \
            xmt[16][UNIT] = x1[16];                                     \
            xmt[17][UNIT] = x1[17];                                     \
            xmt[18][UNIT] = x1[18];                                     \
            xmt[19][UNIT] = x1[19];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 1, MASK);                            \
                                                                        \
        vdsum[0] = spu_add(vdsum[0], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
        x40 = spu_xor(x0_11, XOR3(x1[26], x8_19, x4_15));               \
        x41 = spu_xor(x1_5_9_20, x1[12]);                               \
        x42 = x6_10_17_21_28;                                           \
        x43 = spu_xor(x7_11_22, x3_14_18_29);                           \
        if (UNIT != 4) {                                                \
            xmt[4][UNIT] = x1[4];                                       \
            xmt[5][UNIT] = x1[5];                                       \
            xmt[6][UNIT] = x1[6];                                       \
            xmt[7][UNIT] = x1[7];                                       \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 0, MASK);                             \
                                                                        \
        x40 = XOR3(x1[2], x9_20, x6_13_17_24);                          \
        x41 = spu_xor(x3_28, x17_21);                                   \
        x42 = XOR3(XOR3(x1[26], x8_19, x4_15), x0_18, x22_29);          \
        x43 = spu_xor(x1_5_9_20, x1[23]);                               \
        if (UNIT != 4) {                                                \
            xmt[20][UNIT] = x1[20];                                     \
            xmt[21][UNIT] = x1[21];                                     \
            xmt[22][UNIT] = x1[22];                                     \
            xmt[23][UNIT] = x1[23];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 1, MASK);                            \
                                                                        \
        vdsum[1] = spu_add(vdsum[1], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
        x40 = x8_12_23;                                                 \
        x41 = x9_13_20_24_31;                                           \
        x42 = spu_xor(x6_10, x1[17]);                                   \
        x43 = x0_11;                                                    \
        if (UNIT != 4) {                                                \
            xmt[8][UNIT] = x1[8];                                       \
            xmt[9][UNIT] = x1[9];                                       \
            xmt[10][UNIT] = x1[10];                                     \
            xmt[11][UNIT] = x1[11];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 0, MASK);                             \
                                                                        \
        x40 = XOR3(x13_24, x20_31, x6_10_17_21_28);                     \
        x41 = XOR3(x7_11_22, x18_29, x3_25);                            \
        x42 = spu_xor(x15_26, x8_12_23);                                \
        x43 = XOR3(x1[16], x9_13_20_24_31, x1[27]);                     \
        if (UNIT != 4) {                                                \
            xmt[24][UNIT] = x1[24];                                     \
            xmt[25][UNIT] = x1[25];                                     \
            xmt[26][UNIT] = x1[26];                                     \
            xmt[27][UNIT] = x1[27];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 1, MASK);                            \
                                                                        \
        vdsum[2] = spu_add(vdsum[2], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
                                                                        \
        x40 = spu_xor(x1_12, x8_19);                                    \
        x41 = spu_xor(x2_9_13_17_28, x1[20]);                           \
        x42 = x3_14_18_29;                                              \
        x43 = x4_15;                                                    \
        if (UNIT != 4) {                                                \
            xmt[12][UNIT] = x1[12];                                     \
            xmt[13][UNIT] = x1[13];                                     \
            xmt[14][UNIT] = x1[14];                                     \
            xmt[15][UNIT] = x1[15];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 0, MASK);                            \
                                                                        \
        x40 = x6_10_28;                                                 \
        x41 = spu_xor(x0_11, x18_29);                                   \
        x42 = spu_xor(x8_30, x1_12);                                    \
        x43 = spu_xor(x2_9_13_17_28, x1[31]);                           \
        if (UNIT != 4) {                                                \
            xmt[28][UNIT] = x1[28];                                     \
            xmt[29][UNIT] = x1[29];                                     \
            xmt[30][UNIT] = x1[30];                                     \
            xmt[31][UNIT] = x1[31];                                     \
        }                                                               \
        PRE_ACCUMULATE_LAST(UNIT, 1, MASK);                            \
                                                                        \
        vdsum[3] = spu_add(vdsum[3], (vec_uint4)spu_sumb(vcsum[1], vcsum[0])); \
    }




#define BITSUM(BIT)                                         \
    vd = vdsum[BIT];                                        \
    vd = spu_add(vd, spu_rlqwbyte(vd, 8));                  \
    vd = spu_add(vd, spu_rlqwbyte(vd, 4));                  \
    vsum = spu_add(vsum, spu_sl(vd, (12 - (BIT << 2))));


inline static void
mt2xmt(void)
{
    // xmt[0][*]  mt 32bitΤ
    // MSB (ӥåǥɽǤ0bit) ˤʤ
    int i, j, k;
    for (i = 31; i >= 0; i--) {
        vec_uint4 *vmt = (vec_uint4 *)(void *)mt;
        unsigned n = 0;
        // 512bitܤޤ
        xmt[i][0] = xmt[i][1] = xmt[i][2] = xmt[i][3] = xmt[i][4]
            = spu_splats(0U);

        for (j = 0; j < 4; j++) {
            for (k = 0; k < 128 / 4; k++) {
                xmt[i][j] = spu_xor(spu_rlqw(xmt[i][j], 4),
                                    spu_gather(vmt[n]));
                vmt[n] = spu_rlmaskqw(vmt[n], -1);
                n++;
            }
            xmt[i][j] = spu_rlqwbyte(xmt[i][j], 4);
        }
        // 624bitܤޤǤλĤ
        for (k = 0; k < (624 - 128 * 4) / 4; k++) {
            xmt[i][4] = spu_xor(spu_rlqw(xmt[i][4], 4),
                                spu_gather(vmt[n]));
            vmt[n] = spu_rlmaskqw(vmt[n], -1);
            n++;
        }
        xmt[i][4] = spu_rlqwbyte(xmt[i][j], 6); // 16bitʤΤ
    }
}

static 
unsigned __attribute__((noinline))
update_table_and_genrand_int32_624(int blocks, int rest)
{
    int i;

    vec_uint4 vd;

    volatile vec_uint4 vdsum[4];
    vec_uchar16 vcsum[2];

    vec_uint4 vsum = spu_splats(0U);

    const vec_uint4 vsel1 = (vec_uint4){0x80000000, 0x00000000,
                                        0x00000000, 0x00000000};

    const vec_uint4 vsel112 = (vec_uint4){0x00000000, 0x00000000,
                                          0x00000000, 0x00008000};

    const vec_uint4 vsel13 = (vec_uint4){0xfff80000, 0x00000000,
                                         0x00000000, 0x00000000}; 


    const vec_uint4 vsel29 = (vec_uint4){0xfffffff8, 0x00000000,
                                         0x00000000, 0x00000000};

    vec_uint4 vmask, vmask1 = spu_splats(0xffffffff), vmask2;
    
    vec_uint4 y31;
    vec_uint4 far1;
    vec_uint4 y_next, y_cur;

    vec_uint4 x40, x41, x42, x43;
    vec_uint4 x1[32];

    int rest_shift;

    vec_uint4 x0_11, x4_15, x5_16, x3_10, x0_7, x8_19;
    vec_uint4 x1_12, x18_29, x3_14, x0_18, x20_28, x28_31;
    vec_uint4 x1_5, x6_10, x13_24, x6_17, x17_21, x3_28; 
    vec_uint4 x8_30, x9_20, x15_26, x20_31, x3_25;               
    vec_uint4 x8_12, x22_29, x7_11_22, x2_9_13_17_28;
    vec_uint4 x8_15_19_26, x1_5_9_20, x7_11, x6_13_17_24, x6_10_28;
    vec_uint4 x1_5_16, x9_13_20_24_31;
    vec_uint4 x3_14_18_29, x0_4_15_18, x6_10_17_21_28, x8_12_23;

    vdsum[0] = spu_splats(0U);
    vdsum[1] = spu_splats(0U);
    vdsum[2] = spu_splats(0U);
    vdsum[3] = spu_splats(0U);

    rest_shift = 128 - rest % 128;
    
    vmask2 = spu_slqwbytebc(spu_slqw(vmask1, rest_shift), rest_shift);

    MAKE_Y(0, 31);
    MAKE_Y(0, 0);
    UPDATE_TABLE(0, 0);

    MAKE_Y(0, 1);
    UPDATE_TABLE(0, 1);

    MAKE_Y(0, 2);
    UPDATE_TABLE(0, 2);

    MAKE_Y(0, 3);
    UPDATE_TABLE(0, 3);

    MAKE_Y(0, 4);
    UPDATE_TABLE(0, 4);

    MAKE_Y(0, 5);
    UPDATE_TABLE(0, 5);

    MAKE_Y(0, 6);
    UPDATE_TABLE(0, 6);

    MAKE_Y(0, 7);
    UPDATE_TABLE(0, 7);

    MAKE_Y(0, 8);
    UPDATE_TABLE(0, 8);

    MAKE_Y(0, 9);
    UPDATE_TABLE(0, 9);

    MAKE_Y(0, 10);
    UPDATE_TABLE(0, 10);

    MAKE_Y(0, 11);
    UPDATE_TABLE(0, 11);

    MAKE_Y(0, 12);
    UPDATE_TABLE(0, 12);

    MAKE_Y(0, 13);
    UPDATE_TABLE(0, 13);

    MAKE_Y(0, 14);
    UPDATE_TABLE(0, 14);

    MAKE_Y(0, 15);
    UPDATE_TABLE(0, 15);

    MAKE_Y(0, 16);
    UPDATE_TABLE(0, 16);

    MAKE_Y(0, 17);
    UPDATE_TABLE(0, 17);

    MAKE_Y(0, 18);
    UPDATE_TABLE(0, 18);

    MAKE_Y(0, 19);
    UPDATE_TABLE(0, 19);

    MAKE_Y(0, 20);
    UPDATE_TABLE(0, 20);

    MAKE_Y(0, 21);
    UPDATE_TABLE(0, 21);

    MAKE_Y(0, 22);
    UPDATE_TABLE(0, 22);

    MAKE_Y(0, 23);
    UPDATE_TABLE(0, 23);

    MAKE_Y(0, 24);
    UPDATE_TABLE(0, 24);

    MAKE_Y(0, 25);
    UPDATE_TABLE(0, 25);

    MAKE_Y(0, 26);
    UPDATE_TABLE(0, 26);

    MAKE_Y(0, 27);
    UPDATE_TABLE(0, 27);

    MAKE_Y(0, 28);
    UPDATE_TABLE(0, 28);

    MAKE_Y(0, 29);
    UPDATE_TABLE(0, 29);

    MAKE_Y(0, 30);
    UPDATE_TABLE(0, 30);

    UPDATE_TABLE(0, 31);

    si_lnop();
    si_lnop();

    for (i = 0; i < blocks; i++) {

        GEN_RAND_CALC(0);

        MAKE_Y(1, 31);
        MAKE_Y(1, 0);
        UPDATE_TABLE(1, 0);

        MAKE_Y(1, 1);
        UPDATE_TABLE(1, 1);

        MAKE_Y(1, 2);
        UPDATE_TABLE(1, 2);

        MAKE_Y(1, 3);
        UPDATE_TABLE(1, 3);

        MAKE_Y(1, 4);
        UPDATE_TABLE(1, 4);

        MAKE_Y(1, 5);
        UPDATE_TABLE(1, 5);

        MAKE_Y(1, 6);
        UPDATE_TABLE(1, 6);

        MAKE_Y(1, 7);
        UPDATE_TABLE(1, 7);

        MAKE_Y(1, 8);
        UPDATE_TABLE(1, 8);

        MAKE_Y(1, 9);
        UPDATE_TABLE(1, 9);

        MAKE_Y(1, 10);
        UPDATE_TABLE(1, 10);

        MAKE_Y(1, 11);
        UPDATE_TABLE(1, 11);

        MAKE_Y(1, 12);
        UPDATE_TABLE(1, 12);

        MAKE_Y(1, 13);
        UPDATE_TABLE(1, 13);

        MAKE_Y(1, 14);
        UPDATE_TABLE(1, 14);

        MAKE_Y(1, 15);
        UPDATE_TABLE(1, 15);

        MAKE_Y(1, 16);
        UPDATE_TABLE(1, 16);

        MAKE_Y(1, 17);
        UPDATE_TABLE(1, 17);

        MAKE_Y(1, 18);
        UPDATE_TABLE(1, 18);

        MAKE_Y(1, 19);
        UPDATE_TABLE(1, 19);

        MAKE_Y(1, 20);
        UPDATE_TABLE(1, 20);

        MAKE_Y(1, 21);
        UPDATE_TABLE(1, 21);

        MAKE_Y(1, 22);
        UPDATE_TABLE(1, 22);

        MAKE_Y(1, 23);
        UPDATE_TABLE(1, 23);

        MAKE_Y(1, 24);
        UPDATE_TABLE(1, 24);

        MAKE_Y(1, 25);
        UPDATE_TABLE(1, 25);

        MAKE_Y(1, 26);
        UPDATE_TABLE(1, 26);

        MAKE_Y(1, 27);
        UPDATE_TABLE(1, 27);

        MAKE_Y(1, 28);
        UPDATE_TABLE(1, 28);

        MAKE_Y(1, 29);
        UPDATE_TABLE(1, 29);

        MAKE_Y(1, 30);
        UPDATE_TABLE(1, 30);

        UPDATE_TABLE(1, 31);

        GEN_RAND_CALC(1);


        MAKE_Y(2, 31);
        MAKE_Y(2, 0);
        UPDATE_TABLE(2, 0);

        MAKE_Y(2, 1);
        UPDATE_TABLE(2, 1);

        MAKE_Y(2, 2);
        UPDATE_TABLE(2, 2);

        MAKE_Y(2, 3);
        UPDATE_TABLE(2, 3);

        MAKE_Y(2, 4);
        UPDATE_TABLE(2, 4);

        MAKE_Y(2, 5);
        UPDATE_TABLE(2, 5);

        MAKE_Y(2, 6);
        UPDATE_TABLE(2, 6);

        MAKE_Y(2, 7);
        UPDATE_TABLE(2, 7);

        MAKE_Y(2, 8);
        UPDATE_TABLE(2, 8);

        MAKE_Y(2, 9);
        UPDATE_TABLE(2, 9);

        MAKE_Y(2, 10);
        UPDATE_TABLE(2, 10);

        MAKE_Y(2, 11);
        UPDATE_TABLE(2, 11);

        MAKE_Y(2, 12);
        UPDATE_TABLE(2, 12);

        MAKE_Y(2, 13);
        UPDATE_TABLE(2, 13);

        MAKE_Y(2, 14);
        UPDATE_TABLE(2, 14);

        MAKE_Y(2, 15);
        UPDATE_TABLE(2, 15);

        MAKE_Y(2, 16);
        UPDATE_TABLE(2, 16);

        MAKE_Y(2, 17);
        UPDATE_TABLE(2, 17);

        MAKE_Y(2, 18);
        UPDATE_TABLE(2, 18);

        MAKE_Y(2, 19);
        UPDATE_TABLE(2, 19);

        MAKE_Y(2, 20);
        UPDATE_TABLE(2, 20);

        MAKE_Y(2, 21);
        UPDATE_TABLE(2, 21);

        MAKE_Y(2, 22);
        UPDATE_TABLE(2, 22);

        MAKE_Y(2, 23);
        UPDATE_TABLE(2, 23);

        MAKE_Y(2, 24);
        UPDATE_TABLE(2, 24);

        MAKE_Y(2, 25);
        UPDATE_TABLE(2, 25);

        MAKE_Y(2, 26);
        UPDATE_TABLE(2, 26);

        MAKE_Y(2, 27);
        UPDATE_TABLE(2, 27);

        MAKE_Y(2, 28);
        UPDATE_TABLE(2, 28);

        MAKE_Y(2, 29);
        UPDATE_TABLE(2, 29);

        MAKE_Y(2, 30);
        UPDATE_TABLE(2, 30);

        UPDATE_TABLE(2, 31);

        GEN_RAND_CALC(2);

        MAKE_Y(3, 31);
        MAKE_Y(3, 0);
        UPDATE_TABLE(3, 0);

        MAKE_Y(3, 1);
        UPDATE_TABLE(3, 1);

        MAKE_Y(3, 2);
        UPDATE_TABLE(3, 2);

        MAKE_Y(3, 3);
        UPDATE_TABLE(3, 3);

        MAKE_Y(3, 4);
        UPDATE_TABLE(3, 4);

        MAKE_Y(3, 5);
        UPDATE_TABLE(3, 5);

        MAKE_Y(3, 6);
        UPDATE_TABLE(3, 6);

        MAKE_Y(3, 7);
        UPDATE_TABLE(3, 7);

        MAKE_Y(3, 8);
        UPDATE_TABLE(3, 8);

        MAKE_Y(3, 9);
        UPDATE_TABLE(3, 9);

        MAKE_Y(3, 10);
        UPDATE_TABLE(3, 10);

        MAKE_Y(3, 11);
        UPDATE_TABLE(3, 11);

        MAKE_Y(3, 12);
        UPDATE_TABLE(3, 12);

        MAKE_Y(3, 13);
        UPDATE_TABLE(3, 13);

        MAKE_Y(3, 14);
        UPDATE_TABLE(3, 14);

        MAKE_Y(3, 15);
        UPDATE_TABLE(3, 15);

        MAKE_Y(3, 16);
        UPDATE_TABLE(3, 16);

        MAKE_Y(3, 17);
        UPDATE_TABLE(3, 17);

        MAKE_Y(3, 18);
        UPDATE_TABLE(3, 18);

        MAKE_Y(3, 19);
        UPDATE_TABLE(3, 19);

        MAKE_Y(3, 20);
        UPDATE_TABLE(3, 20);

        MAKE_Y(3, 21);
        UPDATE_TABLE(3, 21);

        MAKE_Y(3, 22);
        UPDATE_TABLE(3, 22);

        MAKE_Y(3, 23);
        UPDATE_TABLE(3, 23);

        MAKE_Y(3, 24);
        UPDATE_TABLE(3, 24);

        MAKE_Y(3, 25);
        UPDATE_TABLE(3, 25);

        MAKE_Y(3, 26);
        UPDATE_TABLE(3, 26);

        MAKE_Y(3, 27);
        UPDATE_TABLE(3, 27);

        MAKE_Y(3, 28);
        UPDATE_TABLE(3, 28);

        MAKE_Y(3, 29);
        UPDATE_TABLE(3, 29);

        MAKE_Y(3, 30);
        UPDATE_TABLE(3, 30);

        UPDATE_TABLE(3, 31);

        GEN_RAND_CALC(3);

        MAKE_Y(4, 31);
        MAKE_Y(4, 0);
        UPDATE_TABLE(4, 0);

        MAKE_Y(4, 1);
        UPDATE_TABLE(4, 1);

        MAKE_Y(4, 2);
        UPDATE_TABLE(4, 2);

        MAKE_Y(4, 3);
        UPDATE_TABLE(4, 3);

        MAKE_Y(4, 4);
        UPDATE_TABLE(4, 4);

        MAKE_Y(4, 5);
        UPDATE_TABLE(4, 5);

        MAKE_Y(4, 6);
        UPDATE_TABLE(4, 6);

        MAKE_Y(4, 7);
        UPDATE_TABLE(4, 7);

        MAKE_Y(4, 8);
        UPDATE_TABLE(4, 8);

        MAKE_Y(4, 9);
        UPDATE_TABLE(4, 9);

        MAKE_Y(4, 10);
        UPDATE_TABLE(4, 10);

        MAKE_Y(4, 11);
        UPDATE_TABLE(4, 11);

        MAKE_Y(4, 12);
        UPDATE_TABLE(4, 12);

        MAKE_Y(4, 13);
        UPDATE_TABLE(4, 13);

        MAKE_Y(4, 14);
        UPDATE_TABLE(4, 14);

        MAKE_Y(4, 15);
        UPDATE_TABLE(4, 15);

        MAKE_Y(4, 16);
        UPDATE_TABLE(4, 16);

        MAKE_Y(4, 17);
        UPDATE_TABLE(4, 17);

        MAKE_Y(4, 18);
        UPDATE_TABLE(4, 18);

        MAKE_Y(4, 19);
        UPDATE_TABLE(4, 19);

        MAKE_Y(4, 20);
        UPDATE_TABLE(4, 20);

        MAKE_Y(4, 21);
        UPDATE_TABLE(4, 21);

        MAKE_Y(4, 22);
        UPDATE_TABLE(4, 22);

        MAKE_Y(4, 23);
        UPDATE_TABLE(4, 23);

        MAKE_Y(4, 24);
        UPDATE_TABLE(4, 24);

        MAKE_Y(4, 25);
        UPDATE_TABLE(4, 25);

        MAKE_Y(4, 26);
        UPDATE_TABLE(4, 26);

        MAKE_Y(4, 27);
        UPDATE_TABLE(4, 27);

        MAKE_Y(4, 28);
        UPDATE_TABLE(4, 28);

        MAKE_Y(4, 29);
        UPDATE_TABLE(4, 29);

        MAKE_Y(4, 30);
        UPDATE_TABLE(4, 30);

        UPDATE_TABLE(4, 31);

        GEN_RAND_CALC(4);

        MAKE_Y(0, 31);
        MAKE_Y(0, 0);
        UPDATE_TABLE(0, 0);

        MAKE_Y(0, 1);
        UPDATE_TABLE(0, 1);

        MAKE_Y(0, 2);
        UPDATE_TABLE(0, 2);

        MAKE_Y(0, 3);
        UPDATE_TABLE(0, 3);

        MAKE_Y(0, 4);
        UPDATE_TABLE(0, 4);

        MAKE_Y(0, 5);
        UPDATE_TABLE(0, 5);

        MAKE_Y(0, 6);
        UPDATE_TABLE(0, 6);

        MAKE_Y(0, 7);
        UPDATE_TABLE(0, 7);

        MAKE_Y(0, 8);
        UPDATE_TABLE(0, 8);

        MAKE_Y(0, 9);
        UPDATE_TABLE(0, 9);

        MAKE_Y(0, 10);
        UPDATE_TABLE(0, 10);

        MAKE_Y(0, 11);
        UPDATE_TABLE(0, 11);

        MAKE_Y(0, 12);
        UPDATE_TABLE(0, 12);

        MAKE_Y(0, 13);
        UPDATE_TABLE(0, 13);

        MAKE_Y(0, 14);
        UPDATE_TABLE(0, 14);

        MAKE_Y(0, 15);
        UPDATE_TABLE(0, 15);

        MAKE_Y(0, 16);
        UPDATE_TABLE(0, 16);

        MAKE_Y(0, 17);
        UPDATE_TABLE(0, 17);

        MAKE_Y(0, 18);
        UPDATE_TABLE(0, 18);

        MAKE_Y(0, 19);
        UPDATE_TABLE(0, 19);

        MAKE_Y(0, 20);
        UPDATE_TABLE(0, 20);

        MAKE_Y(0, 21);
        UPDATE_TABLE(0, 21);

        MAKE_Y(0, 22);
        UPDATE_TABLE(0, 22);

        MAKE_Y(0, 23);
        UPDATE_TABLE(0, 23);

        MAKE_Y(0, 24);
        UPDATE_TABLE(0, 24);

        MAKE_Y(0, 25);
        UPDATE_TABLE(0, 25);

        MAKE_Y(0, 26);
        UPDATE_TABLE(0, 26);

        MAKE_Y(0, 27);
        UPDATE_TABLE(0, 27);

        MAKE_Y(0, 28);
        UPDATE_TABLE(0, 28);

        MAKE_Y(0, 29);
        UPDATE_TABLE(0, 29);

        MAKE_Y(0, 30);
        UPDATE_TABLE(0, 30);

        UPDATE_TABLE(0, 31);


    }

    if (rest >= 128) {
        vmask = vmask1;
        rest -= 128;
    } else if (rest > 0) {
        vmask = vmask2;
        rest = 0;
    } else {
        vmask = spu_splats(0U);
    }


    GEN_RAND_CALC_LAST(0, vmask);

    if (rest >= 128) {
        vmask = vmask1;
        rest -= 128;
    } else if (rest > 0) {
        vmask = vmask2;
        rest = 0;
    } else {
        vmask = spu_splats(0U);
    }



    MAKE_Y(1, 31);
    MAKE_Y(1, 0);
    UPDATE_TABLE(1, 0);

    MAKE_Y(1, 1);
    UPDATE_TABLE(1, 1);

    MAKE_Y(1, 2);
    UPDATE_TABLE(1, 2);

    MAKE_Y(1, 3);
    UPDATE_TABLE(1, 3);

    MAKE_Y(1, 4);
    UPDATE_TABLE(1, 4);

    MAKE_Y(1, 5);
    UPDATE_TABLE(1, 5);

    MAKE_Y(1, 6);
    UPDATE_TABLE(1, 6);

    MAKE_Y(1, 7);
    UPDATE_TABLE(1, 7);

    MAKE_Y(1, 8);
    UPDATE_TABLE(1, 8);

    MAKE_Y(1, 9);
    UPDATE_TABLE(1, 9);

    MAKE_Y(1, 10);
    UPDATE_TABLE(1, 10);

    MAKE_Y(1, 11);
    UPDATE_TABLE(1, 11);

    MAKE_Y(1, 12);
    UPDATE_TABLE(1, 12);

    MAKE_Y(1, 13);
    UPDATE_TABLE(1, 13);

    MAKE_Y(1, 14);
    UPDATE_TABLE(1, 14);

    MAKE_Y(1, 15);
    UPDATE_TABLE(1, 15);

    MAKE_Y(1, 16);
    UPDATE_TABLE(1, 16);

    MAKE_Y(1, 17);
    UPDATE_TABLE(1, 17);

    MAKE_Y(1, 18);
    UPDATE_TABLE(1, 18);

    MAKE_Y(1, 19);
    UPDATE_TABLE(1, 19);

    MAKE_Y(1, 20);
    UPDATE_TABLE(1, 20);

    MAKE_Y(1, 21);
    UPDATE_TABLE(1, 21);

    MAKE_Y(1, 22);
    UPDATE_TABLE(1, 22);

    MAKE_Y(1, 23);
    UPDATE_TABLE(1, 23);

    MAKE_Y(1, 24);
    UPDATE_TABLE(1, 24);

    MAKE_Y(1, 25);
    UPDATE_TABLE(1, 25);

    MAKE_Y(1, 26);
    UPDATE_TABLE(1, 26);

    MAKE_Y(1, 27);
    UPDATE_TABLE(1, 27);

    MAKE_Y(1, 28);
    UPDATE_TABLE(1, 28);

    MAKE_Y(1, 29);
    UPDATE_TABLE(1, 29);

    MAKE_Y(1, 30);
    UPDATE_TABLE(1, 30);

    UPDATE_TABLE(1, 31);

    GEN_RAND_CALC_LAST(1, vmask);

    if (rest >= 128) {
        vmask = vmask1;
        rest -= 128;
    } else if (rest > 0) {
        vmask = vmask2;
        rest = 0;
    } else {
        vmask = spu_splats(0U);
    }


    MAKE_Y(2, 31);
    MAKE_Y(2, 0);
    UPDATE_TABLE(2, 0);

    MAKE_Y(2, 1);
    UPDATE_TABLE(2, 1);

    MAKE_Y(2, 2);
    UPDATE_TABLE(2, 2);

    MAKE_Y(2, 3);
    UPDATE_TABLE(2, 3);

    MAKE_Y(2, 4);
    UPDATE_TABLE(2, 4);

    MAKE_Y(2, 5);
    UPDATE_TABLE(2, 5);

    MAKE_Y(2, 6);
    UPDATE_TABLE(2, 6);

    MAKE_Y(2, 7);
    UPDATE_TABLE(2, 7);

    MAKE_Y(2, 8);
    UPDATE_TABLE(2, 8);

    MAKE_Y(2, 9);
    UPDATE_TABLE(2, 9);

    MAKE_Y(2, 10);
    UPDATE_TABLE(2, 10);

    MAKE_Y(2, 11);
    UPDATE_TABLE(2, 11);

    MAKE_Y(2, 12);
    UPDATE_TABLE(2, 12);

    MAKE_Y(2, 13);
    UPDATE_TABLE(2, 13);

    MAKE_Y(2, 14);
    UPDATE_TABLE(2, 14);

    MAKE_Y(2, 15);
    UPDATE_TABLE(2, 15);

    MAKE_Y(2, 16);
    UPDATE_TABLE(2, 16);

    MAKE_Y(2, 17);
    UPDATE_TABLE(2, 17);

    MAKE_Y(2, 18);
    UPDATE_TABLE(2, 18);

    MAKE_Y(2, 19);
    UPDATE_TABLE(2, 19);

    MAKE_Y(2, 20);
    UPDATE_TABLE(2, 20);

    MAKE_Y(2, 21);
    UPDATE_TABLE(2, 21);

    MAKE_Y(2, 22);
    UPDATE_TABLE(2, 22);

    MAKE_Y(2, 23);
    UPDATE_TABLE(2, 23);

    MAKE_Y(2, 24);
    UPDATE_TABLE(2, 24);

    MAKE_Y(2, 25);
    UPDATE_TABLE(2, 25);

    MAKE_Y(2, 26);
    UPDATE_TABLE(2, 26);

    MAKE_Y(2, 27);
    UPDATE_TABLE(2, 27);

    MAKE_Y(2, 28);
    UPDATE_TABLE(2, 28);

    MAKE_Y(2, 29);
    UPDATE_TABLE(2, 29);

    MAKE_Y(2, 30);
    UPDATE_TABLE(2, 30);

    UPDATE_TABLE(2, 31);

    GEN_RAND_CALC_LAST(2, vmask);

    if (rest >= 128) {
        vmask = vmask1;
        rest -= 128;
    } else if (rest > 0) {
        vmask = vmask2;
        rest = 0;
    } else {
        vmask = spu_splats(0U);
    }


    MAKE_Y(3, 31);
    MAKE_Y(3, 0);
    UPDATE_TABLE(3, 0);

    MAKE_Y(3, 1);
    UPDATE_TABLE(3, 1);

    MAKE_Y(3, 2);
    UPDATE_TABLE(3, 2);

    MAKE_Y(3, 3);
    UPDATE_TABLE(3, 3);

    MAKE_Y(3, 4);
    UPDATE_TABLE(3, 4);

    MAKE_Y(3, 5);
    UPDATE_TABLE(3, 5);

    MAKE_Y(3, 6);
    UPDATE_TABLE(3, 6);

    MAKE_Y(3, 7);
    UPDATE_TABLE(3, 7);

    MAKE_Y(3, 8);
    UPDATE_TABLE(3, 8);

    MAKE_Y(3, 9);
    UPDATE_TABLE(3, 9);

    MAKE_Y(3, 10);
    UPDATE_TABLE(3, 10);

    MAKE_Y(3, 11);
    UPDATE_TABLE(3, 11);

    MAKE_Y(3, 12);
    UPDATE_TABLE(3, 12);

    MAKE_Y(3, 13);
    UPDATE_TABLE(3, 13);

    MAKE_Y(3, 14);
    UPDATE_TABLE(3, 14);

    MAKE_Y(3, 15);
    UPDATE_TABLE(3, 15);

    MAKE_Y(3, 16);
    UPDATE_TABLE(3, 16);

    MAKE_Y(3, 17);
    UPDATE_TABLE(3, 17);

    MAKE_Y(3, 18);
    UPDATE_TABLE(3, 18);

    MAKE_Y(3, 19);
    UPDATE_TABLE(3, 19);

    MAKE_Y(3, 20);
    UPDATE_TABLE(3, 20);

    MAKE_Y(3, 21);
    UPDATE_TABLE(3, 21);

    MAKE_Y(3, 22);
    UPDATE_TABLE(3, 22);

    MAKE_Y(3, 23);
    UPDATE_TABLE(3, 23);

    MAKE_Y(3, 24);
    UPDATE_TABLE(3, 24);

    MAKE_Y(3, 25);
    UPDATE_TABLE(3, 25);

    MAKE_Y(3, 26);
    UPDATE_TABLE(3, 26);

    MAKE_Y(3, 27);
    UPDATE_TABLE(3, 27);

    MAKE_Y(3, 28);
    UPDATE_TABLE(3, 28);

    MAKE_Y(3, 29);
    UPDATE_TABLE(3, 29);

    MAKE_Y(3, 30);
    UPDATE_TABLE(3, 30);

    UPDATE_TABLE(3, 31);

    GEN_RAND_CALC_LAST(3, vmask);

    if (rest > 0) {
        vmask = vmask2;
        rest = 0;
    } else {
        vmask = spu_splats(0U);
    }

    MAKE_Y(4, 31);
    MAKE_Y(4, 0);
    UPDATE_TABLE(4, 0);

    MAKE_Y(4, 1);
    UPDATE_TABLE(4, 1);

    MAKE_Y(4, 2);
    UPDATE_TABLE(4, 2);

    MAKE_Y(4, 3);
    UPDATE_TABLE(4, 3);

    MAKE_Y(4, 4);
    UPDATE_TABLE(4, 4);

    MAKE_Y(4, 5);
    UPDATE_TABLE(4, 5);

    MAKE_Y(4, 6);
    UPDATE_TABLE(4, 6);

    MAKE_Y(4, 7);
    UPDATE_TABLE(4, 7);

    MAKE_Y(4, 8);
    UPDATE_TABLE(4, 8);

    MAKE_Y(4, 9);
    UPDATE_TABLE(4, 9);

    MAKE_Y(4, 10);
    UPDATE_TABLE(4, 10);

    MAKE_Y(4, 11);
    UPDATE_TABLE(4, 11);

    MAKE_Y(4, 12);
    UPDATE_TABLE(4, 12);

    MAKE_Y(4, 13);
    UPDATE_TABLE(4, 13);

    MAKE_Y(4, 14);
    UPDATE_TABLE(4, 14);

    MAKE_Y(4, 15);
    UPDATE_TABLE(4, 15);

    MAKE_Y(4, 16);
    UPDATE_TABLE(4, 16);

    MAKE_Y(4, 17);
    UPDATE_TABLE(4, 17);

    MAKE_Y(4, 18);
    UPDATE_TABLE(4, 18);

    MAKE_Y(4, 19);
    UPDATE_TABLE(4, 19);

    MAKE_Y(4, 20);
    UPDATE_TABLE(4, 20);

    MAKE_Y(4, 21);
    UPDATE_TABLE(4, 21);

    MAKE_Y(4, 22);
    UPDATE_TABLE(4, 22);

    MAKE_Y(4, 23);
    UPDATE_TABLE(4, 23);

    MAKE_Y(4, 24);
    UPDATE_TABLE(4, 24);

    MAKE_Y(4, 25);
    UPDATE_TABLE(4, 25);

    MAKE_Y(4, 26);
    UPDATE_TABLE(4, 26);

    MAKE_Y(4, 27);
    UPDATE_TABLE(4, 27);

    MAKE_Y(4, 28);
    UPDATE_TABLE(4, 28);

    MAKE_Y(4, 29);
    UPDATE_TABLE(4, 29);

    MAKE_Y(4, 30);
    UPDATE_TABLE(4, 30);

    UPDATE_TABLE(4, 31);

    GEN_RAND_CALC_LAST(4, vmask);

    BITSUM(0);
    BITSUM(1);
    BITSUM(2);
    BITSUM(3);

    return spu_extract(vsum, 0);
}


unsigned int
genrand_mine(int num_rand)
{
    int blocks = num_rand / N;
    int rest = num_rand - blocks * N;

    mt2xmt();

    if (__builtin_expect((rest == 0), 0)) {
        blocks--;
        rest = 624;
    }

    return update_table_and_genrand_int32_624(blocks, rest);
}
